/*
LoLo Switcher

The low level keyboard language switcher for X11
Copyright 2011 Sergey M. Stepanov
Contact: xintrea@gmail.com, www.webhamster.ru
Licenses: GPL v.3, BSD
Volgodonsk, 2011
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/input.h>
#include <linux/joystick.h>
#include <sys/ioctl.h>
#include <pthread.h>

#include "Main.h"
#include "Config.h"
#include "EventQueue.h"
#include "KeyLayout.h"

// 0.01 - Первая рабочая версия

// 0.02 - Основана на версии 0.01
//      - Переделан механизм поиска паттерна кодов 
//        с простой маски на регулярное выражение

// 0.03 - Основана на версии 0.02
//      - Сделано кеширование скомпилированных регулярных выражений
//        в первом приближении

// 0.04 - Основана на версии 0.03
//      - Подключение к дисплею вынесено в отдельный метод класса KeyLayout
//      - Изменен механизм циклического переключения раскладок,
//        теперь вместо статической переменной используется
//        определение номера текущей раскладки. Это нужно для
//        одновременной работы с другими менеджерами 
//        управления раскладками

// 0.05 - Основана на версии 0.04
//      - Добавлены команды установки хозяина на создаваемый
//        дефолтный каталог и файл конфигурации. Это необходимо
//        так как LoLo Switcher может быть запущен из-под sudo,
//        после чего обычный пользователь на сможет редактировать
//        файл конфигурации

// 0.06 - Основана на версии 0.05
//      - Сделано схлопывание одинаковых кодов в очереди событий
//      - Сделана блокировка переключения на конкретный язык, 
//        если он уже установлен (при прямом методе переключения)

// 0.07 - Основана на версии 0.06
//      - В конфиг добавлены конфигурирующие параметры CommandX 
//        которые содержат Bash команды, которые срабатывают при
//        переключении на новый язык
//      - Запуск Bash команд сделан в отдельных тредах
//      - В стандартный конфиг добавлены строки с Bash командами
//        которые издают звук на PC Seaker при переключении раскладки

// 0.08 - Основана на версии 0.07
//      - Добавлено создание каталога ~/.config в том случае если его
//        нет в системе         
//      - Добавлены опции показа номера версии

// 0.09 - Основана на версии 0.08
//      - Исправлен механизм определения текущей раскладки и возможности
//        переключения на другую. Теперь нет статической переменной
//        которая следила за номером раскладки, номер раскладки 
//        определяется динамически

// 0.10 - Основана на версии 0.09
//      - Доработано кеширование регулярных выражений согласно
//        документации в pcreprecompile

// 0.11 - Основана на версии 0.10
//      - Сделано правильное подключение к дисплею X11 один раз
//        при старте программы, а не каждый раз при использовании
//        функций X11, которое приводило к ошибке
//        Maximum number of clients reachedError XOpenDisplay
//      - Сделана чистка кода, чтобы небыло предупреждений 
//        о прекращении использования типа char* в строковых C-функциях

// 0.12 - Основана на версии 0.11
//      - При выполнении команд, прописанных в опциях CommandX
//        теперь при запуске LoLo Switcher с установленным SUID, если
//        запуск делается от от обычного пользователя, происходит переключение 
//        эффективного ID на идентификатор обычного пользователя, 
//        который и запустил программу. Это нужно для того, чтобы
//        обычный пользователь не имел возможности выполнять произвольные
//        команды с правами рута

// 0.13 - Основана на версии 0.12
//      - Добавлено считывание переменной EventFilter из конфига
//      - Переделана фильрация событий с константы, прописанной в коде,
//        на проверку регулярного выражения, заданного 
//        в переменной EventFilter

// 0.14 - Основана на версии 0.13
//      - В метод печати теста считывания конфига (по опции -t3)
//        добавлен вывод содержимого переменной EventFilter
//      - В файл CMakeLists.txt добавлены опции инсталляции
//      - Добавлен файл postinstall.cmake, который необходим для
//        правильного завершения инсталляции

// 0.15 - Основана на версии 0.14
//      - Сделана структура универсального события universal_event, чтобы
//        можно было обрабатывать разные типы устройств
//      - В конфиг добавлена переменна DeviceType, указыващая тип
//        устройства - клавиатура или джойстик
//      - Опрос устройства ввода был вынесен в отдельную
//        функцию getUniversalEvents(), в ней реализовано заполнение
//        массива универсальных событий либо событиями клавиатуры,
//        либо событиями джойстика

// 0.16 - Основана на версии 0.15
//      - Исправлена кодировка некоторых *.cpp и *.h файлов,
//        которые имели cp1251 вместо UTF-8
//      - В CMakeLists.txt задана опция компилирования -O0 для
//        файла EventQueue.cpp, так как в нем gcc генерирует
//        неправильный цикл обратного обхода очереди для микропроцессора AMD

// 0.17 - Основана на версии 0.16
//      - В функции отдельного потока threadFunc() добавлено закрытие
//        потока bash-команды с помощью вызова pclose()
//      - Изменен механиз вызова функции отдельного потока. Сделана правильная
//        передача номера выполняемой команды, чтобы небыло потенциальной
//        порчи стека. Добавлена устновка атрибутов потока в
//        значение PTHREAD_CREATE_DETACHED, чтобы ресурсы потока после
//        его завершения были освоождены для использования новыми потоками.

// 0.18 - Основана на версии 0.17
//      - Удален файл postinstall.cmake, его функционал внесен в CMakeLists.txt
//      - В CMakeLists.txt выставлен режим компиляции Release

// 0.19 - Основана на версии 0.18
//      - Программа проверена с помощью valgrind, сделаны небольшие доработки,
//        теперь при работе в любом режиме нет ни одной ошибки 
//        в инструменте проверки Memcheck

// 0.20 - Основана на версии 0.19
//      - Релизная версия. Сделано два варианта: версия с исходниками
//        и версия с скомпилированным бинарником, доработаны readme-файлы


#define RUN_MODE_UNDEFINED       0
#define RUN_AS_PROCESS           1
#define RUN_HELP                 2
#define RUN_VERSION              3
#define RUN_TEST_KEY_CODE        4
#define RUN_TEST_LANGUAGE_LAYOUT 5
#define RUN_TEST_CONFIG_READ     6

// Представление универсального события для клавиатуры и джойстика
struct universal_event 
{
 int a;
 int b;
 int c;
};

// Прототипы функций на случай если вызов идет раньше объявления
void readConfig(void);
void runTestKeyCode(void);
void runTestLanguageLayout(void);
void runTestConfigRead(void);
void runAsProcess(void);
void runHelp(void);
void runVersion(void);
int getUniversalEvents(int fd, struct universal_event *ev);
bool checkEventFilter(char *text);
void switchLayout(int n);
void executeCommand(int n);
static void *threadFunc(void *arg);

// Глобальный объект с конфигурацией программы
Config config;

// Очередь с событиями клавиатуры
EventQueue eventQueue;

// Объект для работы с раскладками клавиатуры
KeyLayout keyLayout;


int main (int argc, char *argv[]) 
{
 // Определение режима работы программы согласно опциям
 int mode=RUN_MODE_UNDEFINED;

 // Если нет опций
 if(argc < 2)
  mode=RUN_AS_PROCESS; // Запуск как процесс 
 else
  {
   // Иначе есть опции

   // printf("argv[0]: %s\n", argv[0]);
   // printf("argv[1]: %s\n", argv[1]);

   // Если нужно показать текст помощи
   if(strcmp(argv[1], "-h")==0 || strcmp(argv[1], "--help")==0)
    mode=RUN_HELP;

   // Если нужно показать номер версии
   if(strcmp(argv[1], "-v")==0 || strcmp(argv[1], "--version")==0)
    mode=RUN_VERSION;

   // Если нужно запустить тестирование кодов клавиатуры
   if(strcmp(argv[1], "-t1")==0)
    mode=RUN_TEST_KEY_CODE;

   // Если нужно запустить тестирование обнаруженных раскладок
   if(strcmp(argv[1], "-t2")==0)
    mode=RUN_TEST_LANGUAGE_LAYOUT;

   if(strcmp(argv[1], "-t3")==0)
    mode=RUN_TEST_CONFIG_READ;
  }

 // Если режим работы программы не был определен
 if(mode==RUN_MODE_UNDEFINED)
  {
   printf("Error program parameters\n");
   return 1;
  }


 if(mode==RUN_AS_PROCESS)
  runAsProcess();

 if(mode==RUN_HELP)
  runHelp();

 if(mode==RUN_VERSION)
  runVersion();

 if(mode==RUN_TEST_KEY_CODE)
  runTestKeyCode();

 if(mode==RUN_TEST_LANGUAGE_LAYOUT)
  runTestLanguageLayout();

 if(mode==RUN_TEST_CONFIG_READ)
  runTestConfigRead();
}


// Считывание конфига в глобальный объект конфига
void readConfig(void)
{
 // Имя конфиг-файла в директории пользователя
 char fileName[250];
 sprintf(fileName, "%s/.config/loloswitcher/config.ini", config.getUserDirectory());
 
 bool configReadResult=false; 

 // Вначале делается попытка считать конфиг из текущей директории
 configReadResult=config.readFile( (char *)"./config.ini");

 // Если в текущей директории конфига нет
 // То делается попытка считать из стандартной директории, отведенной для программы
 if(!configReadResult)
  configReadResult=config.readFile(fileName);

 // Если нигде нет конфига, создается и считывается стандартный конфиг
 if(!configReadResult)
  {
   config.createStandartConfig();
   configReadResult=config.readFile(fileName);
  }

 // Если никаким способом не удалось считать конфиг
 if(!configReadResult)
  {
   printf("Cant read config file.\n");
   exit(1);
  }

 // config.print();
}


// Помощь
void runHelp(void)
{
 runVersion();

 printf("Usage:\n");
 printf("Without parameter - run standart LoLo Switcher as service\n");
 printf("-h or --help - show this help\n");
 printf("-t1 - run event code test\n");
 printf("-t2 - run language layout test\n");
 printf("-t3 - run read config test\n");
 printf("\n");
}


// Номер версии
void runVersion(void)
{
 printf("LoLo Switcher v.%d.%d\n", PROGRAM_VERSION, PROGRAM_SUBVERSION);
 printf("\n");
}


// Запуск теста кодов клавиатуры при нажатии клавиш
void runTestKeyCode(void)
{
 // Чтение конфига
 readConfig();
 
 printf("Event code test.\n");
 if( strlen(config.getEventFilter()) > 0)
  printf("Note! This codes after filtration by EventFilter value: '%s'\n", config.getEventFilter());
 printf("Type CTRL+C for exit.\n");
 printf("\n");


 // Дескриптор файла устройства
 int fd;

 // Открытие устройства ввода
 if((fd = open(config.getInputDevice(), O_RDONLY)) < 0) 
  {
   printf("Couldn't open input device %s, may be permission denied\n", config.getInputDevice());
   exit(1);
  }

 int eventCount=1;


 // Цикл опроса устройства ввода
 while(true) 
  {
   // Массив событий
   struct universal_event ev[64];

   int ev_count=getUniversalEvents(fd, ev); // Считываются события

   // Перебираются события 
   for(int i = 0; i < ev_count; i++) 
    {
     char eventText[250];
     sprintf(eventText, "%d,%d,%d;\n", ev[i].a, ev[i].b, ev[i].c);

     // Применение фильтра событий
     if( checkEventFilter(eventText) ) 
      {
       bool eventShow=true;

       // Не нужно показывать событие отжатия ENTER в первом
       // полученном событии, так как оно остается после нажатия ENTER
       // при запуске программы в консоли
       if(eventCount==1 && ev[i].a==1 && ev[i].b==28 && ev[i].c==0)
        eventShow=false;

       if(eventShow)
        {
         printf("[%d]:\n", eventCount);
         printf("%s\n", eventText);
         printf("\n");

         eventCount++;
        }
      }
    } // Закрылся цикл перебора событий клавиатуры
  
  } // Закрылся бесконечный цикл опроса устройства ввода

}


// Считывание событий с устройства ввода
// fd - дескриптор открытого файла устройства
// ev - указатель на массив событий
int getUniversalEvents(int fd, struct universal_event *ev)
{
 int ev_count;
 
 // Обработка клавиатуры
 if(config.getDeviceType()==CONFIG_DEVICE_TYPE_KEYBOARD)
  {
   // Массив событий клавиатуры
   struct input_event keyboard_ev[64]; 

   // Считываются события
   size_t rb = read(fd, keyboard_ev, sizeof(keyboard_ev));

   // Проверяется результат считывания
   if(rb < (int) sizeof(struct input_event)) 
    {
     printf("Short input keyboard device read\n");
     exit(1);
    }

   ev_count=(int) (rb / sizeof(struct input_event));

   for(int i = 0; i < ev_count; i++) 
    {
     ev[i].a=keyboard_ev[i].type;
     ev[i].b=keyboard_ev[i].code;
     ev[i].c=keyboard_ev[i].value;
    }
  }


 // Обработка джойстика
 if(config.getDeviceType()==CONFIG_DEVICE_TYPE_JOYSTICK)
  {
   // Массив событий джойстика
   struct js_event joystick_ev[64]; 

   // Считываются события
   size_t rb = read(fd, joystick_ev, sizeof(joystick_ev));

   // Проверяется результат считывания
   if(rb < (int) sizeof(struct js_event)) 
    {
     printf("Short input joystick device read\n");
     exit(1);
    }

   ev_count=(int) (rb / sizeof(struct js_event));

   for(int i = 0; i < ev_count; i++) 
    {
     ev[i].a=joystick_ev[i].type;
     ev[i].b=joystick_ev[i].number;
     ev[i].c=joystick_ev[i].value;
    }
  }


 return ev_count;
}


// Проверка кода события на фильтрующее регулярное выражение
// text - код события
bool checkEventFilter(char *text)
{
 // Если фильтр не задан, любое событие всегда проходит
 if(strlen(config.getEventFilter())==0)
  return true;
 
 // Проверка регулярного выражения EventFilter
 int count = 0;
 int ovector[30];
 count=pcre_exec( (pcre *)config.getEventFilterCompile(), NULL, text, strlen(text), 0, NULL, ovector, 30);

 // printf("Count=%d\n", count);

 if(count<=0)
  return false; // Если ничего небыло найдено
 else
  return true; // Что-то было найдено

}


void runTestLanguageLayout(void)
{
 printf("Language layout test.\n");
 printf("Layout found:\n");
 printf("\n");

 keyLayout.print();
}


void runTestConfigRead(void)
{
 printf("Read config test.\n");

 // Загрузка конфига
 readConfig();

 // Печать конфига (конфиг во всей программе доступен как глобальный объект)
 config.print();
}


// Запуск программы в основном режиме
void runAsProcess(void)
{
 // Чтение конфига
 readConfig();

 // Инициализация очереди событий клавиатуры
 eventQueue.init();

 // Дескриптор файла устройства
 int fd;

 // Открытие устройства ввода
 if((fd = open(config.getInputDevice(), O_RDONLY)) < 0) 
  {
   printf("Couldn't open input device %s, may be permission denied\n", config.getInputDevice());
   exit(1);
  }

 // Цикл опроса устройства ввода
 while(true) 
  {
   // Массив событий
   struct universal_event ev[64];

   int ev_count=getUniversalEvents(fd, ev); // Считываются события

   // Перебираются события 
   for(int i = 0; i < ev_count; i++) 
    {
     char eventText[250];
     sprintf(eventText, "%d,%d,%d;\n", ev[i].a, ev[i].b, ev[i].c);

     // Применение фильтра событий
     if( checkEventFilter(eventText) ) 
      {
       /*
       printf(": type, code, value\n");
       printf("%d, %d, %d\n", ev[i].type, ev[i].code, ev[i].value);
       printf("\n");
       */

       // Событие добавляется в очередь
       eventQueue.add(ev[i].a, ev[i].b, ev[i].c);

       // Перебираются раскладки, для которых нужно обнаруживать последовательности
       // событий в очереди
       for(int layoutCount=0; layoutCount<config.getNumberOfLayout(); layoutCount++)
        {
         // printf("RegExp: '%s'\n", config.getSequence(layoutCount));

         // Для циклического переключения раскладок проверяется 
         // только первая последовательность
         if(config.getSwitchMethod()==0 && layoutCount>=1)
          break;
         
         // Проверка последовательности
         if(eventQueue.checkSequence( config.getSequenceCompile(layoutCount) )==true)
          {
           // printf("Detect codes sequence %d\n", layoutCount);

           // Запускается переключение раскладки
           switchLayout(layoutCount);
          }
          
        } // Закрылся цикл перебора раскладок

       
      }
    }
  }

}


// Переключение раскладки
void switchLayout(int n)
{
 // Если нужно циклическое переключение раскладок
 if(config.getSwitchMethod()==CONFIG_SWITCH_METHOD_CYCLIC)
  {
   int number=keyLayout.getLayout();

   number++;

   if(number>=keyLayout.getLayoutNumber())
    number=0;

   keyLayout.setLayout(number);
   executeCommand(number);

   return;
  }

 // Если нужно прямое переключение раскладок
 if(config.getSwitchMethod()==CONFIG_SWITCH_METHOD_DIRECT)
  {
   int number=keyLayout.getLayout();

   if(number!=n)
    {
     keyLayout.setLayout(n);
     executeCommand(n);
    }
  }

}


// Выполнение нужной Bash команды
void executeCommand(int n)
{
 // Если команда не задана
 if(strlen(config.getCommand(n))==0)
  return;

 int keyboardLayerNum=n;

 // Выполнение команды будет происходит в отдельном потоке
 pthread_t thread;

 // Настройка атрибутов потока
 pthread_attr_t threadAttr; 
 pthread_attr_init(&threadAttr); 
 pthread_attr_setdetachstate(&threadAttr, PTHREAD_CREATE_DETACHED); 

 // int result = pthread_create(&thread, NULL, threadFunc, &keyboardLayerNum);
 int result = pthread_create(&thread, &threadAttr, threadFunc, (void*)keyboardLayerNum);
 
 if(result != 0) 
  {
   printf("Creating command thread false. Error: %d\n", result);
   exit(1);
  }

}


static void *threadFunc(void *arg)
{ 
 // int keyboardLayerNum = * (int *) arg;
 int keyboardLayerNum = (int) arg;

 char cmd[250];

 sprintf(cmd, config.getCommand(keyboardLayerNum));

 bool changeAccess=false;
 int uid=getuid();
 int euid=geteuid();

 // Проверяется, работает ли программа в режиме SUID
 if(uid != euid)
  changeAccess=true; // Если установлен бит SUID, надо менять режим доступа

 // Если надо поменять режим доступа, меняется эффективный ID 
 // на ID пользователя, который запустил программу
 if(changeAccess)
  seteuid( uid );

 // printf("Value in thread: getuid()=%d, geteuid()=%d\n", getuid(), geteuid());
 // printf("Run command %s\n", cmd);

 FILE *processPointer; 

 // Команда выполняется
 if( (processPointer=popen(cmd, "r")) == NULL)
  printf("Can't run command: %s\n", cmd);

 // Ожидается завершение процесса с выполняемой командой и процесс завершается
 pclose(processPointer);

 // Если режим доступа был изменен, эффективный UID возвращается в прежнее состояние
 if(changeAccess)
  seteuid( euid );

 // pthread_exit(NULL);

 return NULL;
}

